;************************************************************
;*   Ansteuerung des CO2-Sensors MH-Z19B und Anzeige mit    *
;*			Crystalfontz LCD mit Controller KS0108	        *
;*                       128 x 64 Pixel						*
;************************************************************
;
;       			 ATMega88 
;   				+--+-+--+                        
;	    (!RESET)PC6	|  |_|	|	PC5(ADC5 SCL)                       
;   	   (RXD)PD0	|		|	PC4(ADC4 SDA)                    
;   	   (TXD)PD1	|		|	PC3(ADC3)               
;	      (INT0)PD2 |       |   PC2(ADC2) 	
;	      (INT1)PD3 |       |   PC1(ADC1) 	
;       (XCK T0)PD4 |       |   PC0(ADC0) 	
;	    		VCC |       |   GND
;   			GND |       |   AREF
;  (XTAL1 TOSC1)PB6 |		|	AVcc            
;  (XTAL2 TOSC2)PB7 |		|	PB5(SCK)            
;		    (T1)PD5	|		|	PB4(MISO)               
;		  (AIN0)PD6	|		|	PB3(MOSI OC2)                      
;		  (AIN1)PD7	|		|	PB2(SS OC1B)                      
;	      (ICP1)PB0 |		|	PB1(OC1A)                  
;   				+-------+                           
;
; PB0	I	Eingang PWM CO2
; PB1	I	nc(pullup)
; PB2	I	nc(pullup) 
; PB3	I	nc(pullup), MOSI
; PB4	I	nc(pullup), MISO
; PB5	I	nc(pullup)	
; PB6	I	nc(pullup), XTAL1
; PB7	I	nc(pullup), XTAL2   
; PC0	I	nc(pullup)
; PC1	O	CS2
; PC2	O	CS1 
; PC3	O	R/W
; PC4	O	D/I
; PC5	O	E
; PD0	O	D0
; PD1	O	D1
; PD2	O	D2 
; PD3	O	D3
; PD4	O	D4
; PD5	O	D5
; PD6	O	D6
; PD7	O	D7

.include "m88pdef.inc"

.equ	F_CPU		=8000000

;Register:
.def Counter		=R9
.def Flag_3			=R10
.def tmpdelay		=R14
.def tmpdelay_lang	=R15
.def temp			=R16
.def temp1			=R17
.def temp2			=R18
.def Cnt			=R19
.def tmpdat			=R20
.def Daten			=R21
.def Pages			=R22
.def Flag			=R23
.def Flag_1			=R24
.def Flag_2			=R25

.equ Vorteiler		=256
.equ Periode		=1000000*Vorteiler/F_CPU	;Dauer eines Timertaktes in Mikrosekunden
.equ Range			=2000			;ppm CO2
.equ Minval			=400			;ppm Mindestwert

.equ CS2			=1
.equ CS1			=2
.equ RW				=3
.equ DI				=4
.equ E				=5

.equ LCDC			=PORTC			;Port fr Befehle
.equ LCDDC			=DDRC

.equ LCDD			=PORTD			;Port fr D0-7
.equ LCDPD			=PIND
.equ LCDDD			=DDRD

.equ YSize			=64				;Spalten (2x64=128)
.equ XSize			=8				;8 Pages

;**************************************************************************************
.dseg
Marke2L:	.BYTE 1			;Speicherplatz fr Timer
Marke2H:	.BYTE 1
Marke3L:	.BYTE 1			;Speicherplatz fr Timer
Marke3H:	.BYTE 1
TimeLL:		.BYTE 1			;Low-Wert der Low-Messung in ms
TimeLH:		.BYTE 1			;High-Wert der Low-Messung
TimeHL:		.BYTE 1			;Low-Wert der High-Messung in ms
TimeHH:		.BYTE 1			;High-Wert der High-Messung 
ppm_L:		.BYTE 1			;Low ppm Anzeigewert fr Grafik
ppmL:		.BYTE 1			;Low BCD-Wert ppm
ppmH:		.BYTE 1			;High BCD-Wert ppm

;***************************************************************************************
.eseg
bis48h:		.BYTE 100		;jede halbe Stunde wird der Wert gespeichert

;***************************************************************************************
.cseg
.org 0
	rjmp Reset
.org ICP1addr
	rjmp IN_Capt
.org OVF0addr
	rjmp Tim_Over

; CPU Register Initialisierung
;****************************************************************************
Reset:
	ldi		temp, low(RAMEND)
	out		SPL, temp			;setup stack pointer
	ldi		temp, high(RAMEND)
	out		SPH, temp			;setup stack pointer

;Ports konfigurieren
	ldi		temp, 0b11111110 
	out		PortB, temp			;alles Eingang mit pullup auer PWM Eingang			

	ldi		temp, 0b111110		;alles Ausgang auer 0
	out		LCDDC, temp
	ldi		temp, 0b000001		;PORTC0 Eingang mit pullup
	out		LCDC, temp		

	ldi		temp, 0xFF			;PORTD alles Ausgang
	out		LCDDD, temp

	ldi		temp, (1<<CS02)|(1<<CS00)	;Timer 0 auf prescaler 1024, berlauf alle 1/8.000.000*1024*256 = 32,768 ms
	out		TCCR0B, temp
	ldi		temp, (1<<TOIE0)	;Interrupt enabled
	sts		TIMSK0, temp

	ldi		temp, (1<<ICNC1)|(0<<ICES1)|(1<<CS12)	;Timer 1 auf Input Capture fallende Flanke und prescaler 256
	sts		TCCR1B, temp
	ldi		temp, (1<<ICIE1)	;Interrupt enabled
	sts		TIMSK1, temp
	
	sei							;Interrupt freigeben			
	
	ldi		temp, 9
	mov		Counter, temp
	clr		temp

	rcall	LCD_init
	rcall	delay500ms
	rcall	LCD_init
	rcall	Bild_Aus
	rcall	Grafik
	rcall	EE_Read_L_loop		;damit Grafik sofort angezeigt wird
	rjmp	Start	
Loop:
rjmp	Loop
;*************************************************************************************************************
;Interrupt zur Erfassung der PWM Werte des Sensors
IN_Capt:						
;mit steigender Flanke wird der Timer gestartet
	push	temp
	push	temp1
	in		temp1, SREG
	cpi		Flag, 1				;Flag H, L gesetzt?
	brsh	_IN_Capt
	clr		temp
	sts		TCNT1H, temp		;Timerwerte lschen
	sts		TCNT1L, temp
	ldi		temp, (1<<ICNC1)|(1<<ICES1)|(1<<CS12)	;Timer 1 auf Input Capture steigende Flanke und prescaler 256
	sts		TCCR1B, temp
	inc		Flag
	out		SREG, temp1
	pop		temp1
	pop		temp
reti
_IN_Capt:						;Timerwert fr steigende Flanke 
	cpi		Flag, 2				;Flag H, L gesetzt?
	breq	__IN_Capt
	lds		temp, ICR1L			;Timerwert einlesen
	sts		Marke2L, temp		;und speichern
	lds		temp, ICR1H			
	sts		Marke2H, temp		 
	ldi		temp, (1<<ICNC1)|(0<<ICES1)|(1<<CS12)	;Timer 1 auf Input Capture fallende Flanke und prescaler 256
	sts		TCCR1B, temp
	inc		Flag
	out		SREG, temp1
	pop		temp1
	pop		temp
reti
__IN_Capt:						;Timerwert fr fallende Flanke
	lds		temp, ICR1L			;Timerwert einlesen
	sts		Marke3L, temp		;und speichern
	lds		temp, ICR1H			
	sts		Marke3H, temp		 
	clr		Flag
	out		SREG, temp1
	pop		temp1
	pop		temp
reti

;***********************************************************************************************
;jede halbe Stunde wird ein ppm_L Wert gespeichert
Tim_Over:
	push	temp1
	in		temp1, SREG
	inc		Flag_1
	cpi		Flag_1, 230					;32,768 ms * 230 = 7,53664 s
	breq	_Tim_Over
	out		SREG, temp1
	pop		temp1
reti
_Tim_Over:
	clr		Flag_1
	inc		Flag_2
	cpi		Flag_2, 235					;7,53664 * 235 = 1771,1104 s /60 ~ 29,52 min
	breq	Wert
	out		SREG, temp1
	pop		temp1
reti
Wert:
	sbic	EECR, EEPE
	rjmp	Wert
	push	temp
	clr		Flag_2
	inc		Flag_3
	lds		temp, ppm_L
	ldi		ZL, low(bis48h)				;Zeiger auf EEPROM Adresse
	ldi		ZH, high(bis48h)
	add		ZL, Counter					;Zhler addieren
	adc		ZH, R8
	out		EEARH, ZH					;Adresse ins EEPROM Adressregister
	out		EEARL, ZL
	out		EEDR, temp					;Daten ins Datenregister schreiben
	sbi		EECR, EEMPE					;EEMPE auf 1 setzen
	sbi		EECR, EEPE					;ppm_L schreiben
	inc		Counter						;Zhler erhhen
	mov		temp, Counter
	cpi		temp, 99					;max 99 Werte werden angezeigt
	breq	_Count
	pop		temp
	out		SREG, temp1
	pop		temp1
reti
_Count:
	clr		Counter
	pop		temp
	out		SREG, temp1
	pop		temp1
reti

;*********************************************************************************************
Start:							;Timerwert steigende Flanke verarbeiten
	lds		temp, Marke2H
	tst		temp
	breq	Start
	lds		temp, Marke2L
	lds		temp1, Marke2H
	rcall	Multi				;Multiplikation mit Periode > Mikrosekunden
	rcall	Div24u				;Division durch 1000 > ms
	sts		TimeLL, r3			;Ergebnis speichern
	sts		TimeLH, r4		
	clr		temp				;Speicher lschen
	sts		Marke2H, temp
	sts		Marke2L, temp
_Start:							;Timerwert fallende Flanke verarbeiten
	lds		temp, Marke3H
	tst		temp
	breq	_Start
	lds		temp, Marke3L
	lds		temp1, Marke3H
	rcall	Multi				;Multiplikation mit Periode
	rcall	Div24u
	sts		TimeHL, r3
	sts		TimeHH, r4
	clr		temp				;Speicher lschen
	sts		Marke3H, temp
	sts		Marke3L, temp

	rcall	ppm					;Umrechnung in ppm
Zurueck:
	rcall	hexToBcd			;Umwandlung in BCD

Ausgabe:
	lds		temp1, ppmH
	swap	temp1
	andi	temp1, 0b00001111
	rcall	Ziff_T
	lds		temp1, ppmH
	andi	temp1, 0b00001111
	rcall	Ziff_H
	lds		temp1, ppmL
	swap	temp1
	andi	temp1, 0b00001111
	rcall	Ziff_Z
	lds		temp1, ppmL
	andi	temp1, 0b00001111
	rcall	Ziff_E
;Ausgabe der Kurvenwerte
	rcall	EE_Read_L
	rcall	long_delay
rjmp	Start

;*******************************************************************************************
delay2us:
	push	temp1
    ldi		temp1, 0x02					;17 Takte ~ 2us
delay2us_:
	dec		temp1
    brne	delay2us_
	pop		temp1
ret

delay5us:
	push	temp1
    ldi		temp1, 0x9					;38 Takte ~ 5us
delay5us_:
	dec		temp1
    brne	delay5us_
	pop		temp1
ret

delay50us:
	push	temp1
    ldi		temp1, 0x85					;406 Takte = 50us
delay50us_:
	dec		temp1
    brne	delay50us_
	pop		temp1
ret

delay75us:
	push	temp1
    ldi		temp1, 0xC9					;614 Takte = 77us
delay75us_:
	dec		temp1
    brne	delay75us_
	pop		temp1
ret

delay100us:
	rcall	delay50us
	rcall	delay50us
ret

delay50ms:								;395.816 Takte, ca. 50ms
	push	temp1
	ldi		temp1, 3
	mov		tmpdelay_lang, temp1
rjmp	long_long	

delay100ms:								;791590 Takte, ca. 100ms
	push	temp1
	ldi		temp1, 5
	mov		tmpdelay_lang, temp1
rjmp	long_long

delay500ms:								;3.957.874 Takte, ca. 500ms
	push	temp1
	ldi		temp1, 21
	mov		tmpdelay_lang, temp1
rjmp	long_long

delay1s:
	rcall	delay500ms
	rcall	delay500ms
ret

long_delay:
	push	temp1
	clr		temp1
	mov		tmpdelay_lang, temp1
long_long:
	ldi		tmpdat, 0
	dec		tmpdelay_lang
	brne	long_delay_loop
	pop		temp1
ret
long:
	clr		tmpdelay
	dec		tmpdat
	brne	long_delay_loop
rjmp	long_long
long_delay_loop:
	dec		tmpdelay
	brne	long_delay_loop
rjmp	long	



	.include "Mathe.inc"
	.include "KS0108_LCD.inc"
	.include "Grafik.inc"

